#!/usr/bin/python2
# -*- coding: UTF-8 -*-

from __future__ import division, absolute_import, print_function

"""
My first attempt to display/analyze and disassemble Enteprise-128 EXOS
file formats / programs (as an UTF-8 text or HTML page).
(C)2012,2013 LGB Gábor Lénárt lgb@lgb.hu
Can be used/distributed/modified according to license of GNU/GPL v3 or later.
More information: http://www.gnu.org/licenses/gpl.html

Nice information sources:
	http://ep128.hu/Ep_Konyv/Tippek.htm
	http://ep128.hu/Ep_Konyv/Gepi_kod_2.htm#342
	http://ep128.hu/Ep_Util/Brd.htm
	http://enterpriseforever.com/

Thanks for the tips/help on the Enterprise Forever forum & especially Zozo
"""

import sys

def die(msg):
	sys.stderr.write("EXIT ON FATAL: " + msg + "\n")
	sys.exit(1)

try:
	from z80disasm import Z80Disassembler, Z80OpcodeMatrixDump
except ImportError as a:
	die("Cannot import from z80disasm.py (it's needed!): " + str(a))



class EP128FileLister(Z80Disassembler):

	PROJECT_NAME = "Enterprise-128 EXOS file analyzer, lister and disassembler"
	VERSION = "0.5.8"
	COPYRIGHT_YEARS = "2012,2013"
	OPTIONS = {
		"html":	"HTML output instead of text",
		"info": "Display some info on the source file too",
		"cset": "Select EP ASCII table, eg.: -cset=NameOfCSet",
		"hex" : "Also show hexdump before each line",
		"out":  "Use file as output instead of stdout, eg: -out=file.txt",
		"dasm": "Do disassembly in case of type-5 prg (UNFINISHED!) otherwise only hex dump",
		"bin": "Treat input as BINARY file (not EXOS!) with LOAD loading and START starting (in hex!) addresses: -bin=LOAD,START (you can give more START addresses with additional commas)",
		"nolinks": "Disables anchor/link creation in HTML mode",
		"png":  "Mimic an EP-128 screen, creates PNG image",
		"toep": "Instead of listing, do a PC->EP text _only_ conversion",
		"savepic": "In case of an image, save them as PC viewable file. Parameter (eg: -savepic=example) is needed to name the base name (without extension) and/or path for the file. Extension will be appended automatically!"
	}
	opt = {}
	tokens = [
		"ALLOCATE", "ASK", "AUTO", "CALL", "CAPTURE", "CASE", "CAUSE", "CLEAR", "CLOSE", "CODE", "CONTINUE", "COPY",
		"DATA", "DEF", "DEF", "DELETE", "DIM", "DISPLAY", "DO", "CHAIN", "EDIT", "ELSE", "ELSE IF", "END",
		"END DEF", "END HANDLER", "END IF", "END SELECT", "END WHEN", "ENVELOPE", "EXIT", "FOR", "GOSUB", "GOTO",
		"GRAPHICS", "HANDLER", "IMAGE", "IF", "IF", "INPUT", "LET", "LINE", "LIST", "LOAD", "LOOP", "MERGE", "NEW",
		"NEXT", "NUMERIC", "OPEN", "OPTION", "OK", "OUT", "PLOT", "POKE", "SPOKE", "PRINT", "PROGRAM", "RANDOMIZE",
		"READ", "REDIRECT", "REM", "RENUMBER", "RESTORE", "RETRY", "RETURN", "RUN", "SAVE", "SELECT", "SET", "SOUND",
		"START", "STOP", "INFO", "STRING", "TEXT", "TOGGLE", "TRACE", "TYPE", "VERIFY", "WHEN", "!", "LLIST", "LPRINT",
		"EXT", "GET", "FLUSH", "LOOK", "PING", "DATE", "TIME", "WAIT", "ON"
	] #: BASIC tokens. Be careful, indices of the tokens within the python list are important!
	untokenized = ["REM", "!", "IMAGE", "DATA"] #: rest of the line is untokenized after these tokens
	#untokenized = [tokens.index(a) for a in ["REM","!","IMAGE","DATA"]]
	signs = [	# TODO: the first (0th ...) element and the ones marked with "?" ...
		" ", "!","\"", "#", "$", "%", "&", "'",
		"(", ")", "*", "+", ",", "-", ".", "/",
		":", ";", "<", "=", ">","<>","<=",">=",
		"[","\\", "]", "^", "?", "?", "?", "?"
	]
	forth_default_vlist = set([
		"ASSEMBLER", "FORTH", "XOR", "U<", "UM/MOD", "UM*", "TRUE", "SWAP", "SP!", "RP!", "ROT", "ROLL", "R@", "R>",
		"PICK", "OVER", "OR", "NOT", "NOR", "NOOP", "NEGATE", "NAND", "MOD", "MIN", "MAX", "J", "I", "FALSE", "DUP",
		"DROP", "C@", "C!", "?BRANCH", "BRANCH", "AND", "ABS", "@", "?DUP", ">R", ">", "=", "<>", "<", "3", "2/", "2-",
		"2+", "2*", "2", "1-!", "1-", "1+!", "1+", "1", "0>", "0=", "0<", "0!", "0", "/MOD", "/", "-ROT", "-3", "-2",
		"-1", "-!", "-", "+!", "+", "*/MOD", "*/", "*", "!", "WORD", "WIPE", "VOC-LINK", "VLIST", "VERSION", "U.", "TIB",
		"THRU", "STATE", "SPAN", "SLOW", "SIGN", "SCR", "S0", "S->D", "RND", "RECURSE", "R0", "R#", "QUIT", "PREV",
		"PAD", "P@", "P!", "OUT", "OCTAL", "NUMBER", "LOAD", "LIST", "LAST", "KEY", "INTERPRET", "INDEX", ">IN", ">BODY",
		"HLD", "HOLD", "HEX", "HERE", "FORTH-83", "FORGET", "FLD", "FIND", "FILL", "FENCE", "FAST", "EXIT", "EXECUTE",
		"ERROR", "ERASE", "EMPTY", "DPL", "DP", "DEPTH", "DELETE", "DEFINITIONS", "DECIMAL", "CURRENT", "CSP", "COUNT",
		"COPY", "CONVERT", "CONTEXT", "CMOVE>", "CMOVE", "BLK", "BLANK", "BINARY", "BASE", "BACKGROUND", "ABORT",
		"?TERMINAL", "?STACK", "?CSP", "?", "<#", ";S", ".S", ".R", ".(", ".", "(SWI)", "(BACKGROUND)", "(ABORT)", "(",
		"'", "-TRAILING", "-->", "#TIB", "#S", "#>", "#", "!CSP", "]", "[COMPILE]", "[']", "[", "WHILE", "VOCABULARY",
		"VARIABLE", "USES", "UNTIL", "THEN", "SMUDGE", "REPEAT", "LOOP", "LITERAL", "LEAVE", "IMMEDIATE", "IF", "FORWARD",
		"ELSE", "DOES>", "DO", "CREATE", "CONSTANT", "COMPILE", "CODE", "C,", "BEGIN", "ASCII", "ALLOT", "ABORT\"",
		">RESOLVE", ">MARK", "<RESOLVE", "<MARK", ";CODE", ";", ":", ".\"", ",", "+LOOP", "SLICE", "CONCAT", "$VARIABLE",
		"$CONSTANT", "$<>", "$.", "$@", "$!", "\"", "INFO", "FREE", "EDIT", "***", "ok", "TIME!", "DATE!", "TIME@",
		"DATE@", "CHARACTER", "PALETTE", "GET", "PUT", "JOY", "OPENOUT", "OPENIN", "RGB", "SPFUNC", "#STATUS", "VADDR",
		"S@", "S!", "FONT", "FKEY", "EXT", "COLOR", "COLOUR", "REDIRECT", "CAPTURE", "ATTRIBUTES", "#KEYBOARD",
		"#EDITOR", "#TEXT", "#GRAPHICS", "REM2", "REM1", "BIAS", "BORDER", "STATUS", "Y", "X", "COLORS", "COLOURS",
		"MODE", "FORMAT", "BAUD", "SPEAKER", "STOP", "TIMER", "BUFFERS", "WHITE", "MAGENTA", "CYAN", "YELLOW", "GREEN",
		"BLUE", "RED", "BLACK", "AT", "DISPLAY", "TOGGLE", "OFF", "ON", "ASK", "SET", "E@", "E!", "CLOSE", "STYLE",
		"+PLOT", "PLOT", "PAPER", "PAINT", "+MOVE", "MOVE", "PLOTTING", "INK", "ELLIPSE", "+DRAW", "DRAW", "SOUND",
		"INTERRUPT", "SYNC", "SOURCE", "EFFECTS", "RIGHT", "LEFT", "DURATION", "PITCH", "ENVELOPE", "RELEASE", "PHASE",
		"CLG", "CLS", "GRAPHICS", "HIRES", "LORES", "ATTRIBUTE", "TEXT", "-TEXT", "DU<", "DU.", "DNEGATE", "DMIN", "DMAX",
		"DABS", "D=", "D<", "D2/", "D0=", "D.R", "D.", "D-", "D+", "D*", "2VARIABLE", "2SWAP", "2ROT", "2OVER", "2LITERAL",
		"2DUP", "2DROP", "2CONSTANT", "2@", "2!", "-2ROT", "UPDATE", "TYPE", "SPACES", "SPACE", "SAVE-BUFFERS", "NAME",
		"LOAD-BUFFERS", "SYSLOAD", "FLUSH", "EXPECT", "EMPTY-BUFFERS", "EMIT", "DEVICE", "DESTROY-BUFFERS",
		"CREATE-BUFFERS", "CR", "BUFFER", "BLOCK"
	]) #: default is-forth "keywords" (upper case!)
	ascii_default = "UK"  # which ASCII table should be used by default
	# Source: http://ep128.hu/Ep_Util/Brd.htm
	ascii = { # patch "ascii_topatch" with these, key: char to replace in ascii_topatch, value: char to replace with
		"UK":	{}, # must be empty patchlist if ascii_topatch is the UK tab exactly!
		"BRD":	{
			35: "#", 64: "§", 91: "Ä", 92: "Ö", 93: "Ü",
			96: "´", # 96: TODO: what is the diff between UK/BRD?
			123: "ä", 124: "ö", 125: "ü", 126: "ß"
		},
		"HUN":	{
			35: "#", 64: "Á", 91: "É", 92: "Ö", 93: "Ü",
			96: "á", 123: "é", 124: "ö", 125: "ü",
			126: "ó"
		},
		"EPPHUN": {
			127: "?", # Ft symbol?
			128: "Á", 129: "É", 130: "í", 131: "ó", 132: "ö",
			133: "ő", 134: "ú", 135: "ü", 136: "ű", 137: "┌",
			138: "└", 139: "│", 140: "├", 141: "┬", 143: "┼",
			144: "á", 145: "é", 146: "í", 147: "ó", 148: "ö",
			149: "ő", 150: "ú", 151: "ü", 152: "ű", 153: "┐",
			154: "┘", 155: "─", 156: "┤", 157: "┴",
			159: "?" # What is this??
		},
		"HFONT":{
			# Note: I cannot find "triple" box drawing chars in unicode ...
			129: "á", 130: "╝", 131: "╚", 132: "í", 133: "Í",
			134: "ó", 135: "ö", 136: "ő", 137: "Ü", 138: "ú",
			139: "ü", 140: "ű", 141: "║",
			142: "?", # I have no idea what it is, "Ft" symbol for (Hungarian) forint?
			143: "Ű", 145: "Á", 146: "Ó", 147: "é", 148: "Ö",
			149: "Ú", 150: "═", 151: "É", 152: "╗", 153: "Ő",
			154: "═", 156: "╔", 157: "║"
		}
	}
	chsetnamemaxlen = max([len(a) for a in ascii])
	# our base charset is UK, so patchlist above for UK should be empty
	# it's expected to only have characters between 32-159 here!!
	# Yes, it would be "nicer" solution to generate 'std' ascii codes and
	# only rewrite what is different on EP, but it's better for me
	# now to show the charset in one table. The "diff" method
	# is used for more csets though!
	ascii_topatch = [
		" ", "!", "\"","£", "$", "%",	#  32 -  37
		"&", "'", "(", ")", "*", "+",	#  38 -  43
		",", "-", ".", "/", "0", "1",	#  44 -  49
		"2", "3", "4", "5", "6", "7",	#  50 -  55
		"8", "9", ":", ";", "<", "=",	#  56 -  61
		">", "?", "@", "A", "B", "C",	#  62 -  67
		"D", "E", "F", "G", "H", "I",	#  68 -  73
		"J", "K", "L", "M", "N", "O",	#  74 -  79
		"P", "Q", "R", "S", "T", "U",	#  80 -  85
		"V", "W", "X", "Y", "Z", "[",	#  86 -  91
		"\\","]", "^", "_", "`", "a",	#  92 -  97
		"b", "c", "d", "e", "f", "g",	#  98 - 103
		"h", "i", "j", "k", "l", "m",	# 104 - 109
		"n", "o", "p", "q", "r", "s",	# 110 - 115
		"t", "u", "v", "w", "x", "y",	# 116 - 121
		"z", "{", "|", "}", "~", "?",	# 122 - 127, here the last char is unknown for me ...
		"©", "#", "Ä", "Ö", "Ü", "Å",	# 128 - 133
		"Ø", "Æ", "é", "Ç", "à", "↑",	# 134 - 139
		"→", "Ω", "▮", "ƒ", "π", "Σ",	# 140 - 145
		"ä", "ö", "ü", "å", "ø", "æ",	# 146 - 151
		"è", "§", "☀", "↓", "←", "μ",	# 152 - 157, char "☀", really is it? on EP it looks like a little man :)
		"←", "■"			# 158 - 159, arrow is duplicated? it's in the previous line as well ...
	]
	# TODO: create some nice stylesheet instead of this "random" and ugly one :(
	html_head = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
	<html>
	<head>
		<META HTTP-EQUIV="Content-Type" CONTENT="text/HTML; charset=UTF-8">
		<title>""" + PROJECT_NAME + """</title>
		<style type="text/css">
			BODY {
	        		background: #000000;
				color: #80FF80;
	        		margin: 2px;
	        		padding: 2px;
			}
			.linehead {
				background: #204020;
			}
			.sign {
				font-weight: bold;
			}
			a:hover {
				background: #FF0000;
				text-decoration: none;
			}
			.string {
				background: #408040;
			}
			.comment {
				color: #A0A0A0;
			}
			.token {
				color: #FF8080;
				font-weight: bold;
			}
			.number {
				color: #C0C0C0;
			}
			.name {
				background: #808080;
			}
			.stat {
				font-weight: bold;
				border: 1px solid #00FF00;
				padding: 10px;
			}
			.hex {
				color: #8080FF;
				font-weight: bold;
			}
		</style>
	</head>
	<body>\n"""
	html_listing_head = "<h1><a href=\"http://epbas.lgb.hu/\">epbas.lgb.hu</a> :: " + PROJECT_NAME + "</h1>\n"
	html_tail = "<br><br><hr>" + PROJECT_NAME + " v" + VERSION + " &copy;" + COPYRIGHT_YEARS + " \"LGB\" Gábor Lénárt (lgb at-sign lgb dot hu)</body></html>"

	def dump_db(self):

		o = "Dumping IS-BASIC tokens:\n\n"
		for a in range(len(self.tokens)):
			o += str(a) + "\t" + self.tokens[a]
			if self.tokens[a] in self.untokenized: o += "\t\t[UNTOK_LEFT]"
			o += "\n"
		o += '\nDefault IS-FORTH vlist:\n'
		n = 8888
		for a in sorted(self.forth_default_vlist):
			if n > 80:
				n = 0
				o += '\n'
			else:
				n += len(a) + 1
				o += ' '
			o += a
		o += '\n\nSpecial signs table for IS-BASIC: "' + '","'.join(self.signs) + '"\n'
		o += "\nAvailable EP-ASCII tables: " + ", ".join(sorted(self.ascii)) + " [default: " + self.ascii_default + "]\n"
		for a in sorted(self.ascii):
			o += "ASCII table \"" + a + "\": " + (self.chsetnamemaxlen - len(a)) * " " + "".join(self.get_ascii_table(a)[32:160]) + "\n"
		o += "\nHandled EXOS file types:\n\n"
		for a in sorted(self.known_exostypes):
			o += str(a) + "\t" + self.known_exostypes[a][0] + ((" " + str(self.known_exostypes[a][2])) if self.known_exostypes[a][2] else "") + "\n"
		o += "\nEnterprise EXOS functions:\n\n"
		for a in sorted(self.EXOS_FUNCS):
			o += hex(a) + "\t" + self.EXOS_FUNCS[a][0] + "\n"
		o += "\nEnterprise IO ports:\n\n"
		for a in sorted(self.EP_IOREGS):
			o += hex(a) + "\t" + self.EP_IOREGS[a] + "\n"
		return o

	def get_unicode_pos(self, c):
		"""This is needed to cover the python2/python3 difference issue. On python3 str is unicode, while it's not with python2."""
		try: return ord(c)
		except TypeError: return ord(c.decode("utf-8"))

	def dump_z80opcodes(self):

		tab = Z80OpcodeMatrixDump().dump()
		o = self.html_head + "<h1><a href=\"http://epbas.lgb.hu/\">epbas.lgb.hu</a> :: Z80 CPU Opcodes</h1>\n\n"
		o += '<table border="0" bgcolor="#204020" cellpadding="3" cellspacing="3">\n<tr bgcolor="#000000">\n'
		for a in ("", "-", "DD", "CB", "DD CB", "ED"): o += '\t<td class="token">' + a + "</td>\n"
		o += "</tr>"
		byte = 0
		for a in tab:
			o += '<tr bgcolor="#000000">\n\t<td class="hex"><code>' + self.hexbyte(byte) + "</code></td>\n"
			byte += 1
			bo = ""
			for b in a:
				if b[0] in ("PREFIX", "UNDEF"): c = ' class="token"'
				else: c = ""
				b = b[0] + "&nbsp;" * (4 - len(b[0])) + " " + b[1] + ('<br><font size="-2"><i>term</i></font>' if b[2] else "")
				if b == bo: b = ""
				o += "\t<td" + c + "><b><code>" + b + "</code></td>\n"
				bo = b
			o += "</tr>"
		return o + "\n</table>" + self.html_tail


	def dump_chset(self):

		o = self.html_head + "<h1><a href=\"http://epbas.lgb.hu/\">epbas.lgb.hu</a> :: Enterprise Character Sets As Unicode Representation</h1>\n" + "<table border=\"0\" bgcolor=\"#204020\" cellpadding=\"3\" cellspacing=\"3\">\n"
		c = {a: self.get_ascii_table(a) for a in self.ascii}
		l = sorted(c)
		for a in range(32, 160):
			if a % 16 == 0:
				o += "<tr bgcolor=\"#000000\"><td colspan=\"2\">&nbsp;</td><td class=\"token\" align=\"center\">" + "</td><td class=\"token\" align=\"center\">".join(l) + "</td></tr>\n"
			o += "<tr bgcolor=\"#000000\"><td align=\"right\" class=\"hex\">" + str(a) + "</td><td class=\"hex\" align=\"right\">" + hex(a)[2:].upper() + "h</td>"
			for b in l:
				o += "<td align=\"center\" bgcolor=\"#" + ("204020" if c[b][a] != c["UK"][a] else "000000") + "\"><font size=\"+2\"><b>"
				o += c[b][a] + "</b></font><br><font size=\"-1\"><i>"
				o += hex(self.get_unicode_pos(c[b][a])).upper().replace("0X", "U+")
				#try:
				#	o += hex(ord(c[b][a])).upper().replace("0X", "U+")
				#except TypeError:
				#	o += hex(ord(c[b][a].decode("utf-8"))).upper().replace("0X", "U+")
				o += "</i></font></td>"
			o += "</tr>\n"
		return o + "</table>\n" + self.html_tail

	def get_ascii_table(self, cset):
		"""This method patches the ascii_topatch with the selected patchlist given as "EP character set"."""
		#return [(self.ascii[cset][a] if a in self.ascii[cset] else a) for a in [" "] * 32 + self.ascii_topatch]
		try:
			if len(self.ascii_topatch) != 128:
				raise RuntimeError("ascii_topatch table should be 128 elements long, but it is " + str(len(self.ascii_topatch)))
			o = [(self.ascii[cset][a] if a in self.ascii[cset] else self.ascii_topatch[a - 32]) for a in range(32, 160)]
			return [" "] * 32 + o + o[:96]
		except IndexError:
			raise RuntimeError("ASCII setup problem")

	def set_options(self, options):

		for a in options:
			a, v = (a.split("=", 1) + [""])[:2]
			if a[0] != "-": raise RuntimeError("Bad option syntax: " + a)
			if a[1:] not in self.OPTIONS: raise RuntimeError("Unknown option: " + a)
			self.opt[a[1:]] = v
		if "cset" not in self.opt: self.opt["cset"] = self.ascii_default
		if "out" not in self.opt: self.opt["out"] = ""
		if "bin" in self.opt:
			a = self.opt["bin"].split(",")
			if len(a) < 2:
				raise RuntimeError("With -bin you must use the syntax -bin=LOAD,ADDR[,ADDR2[...]]")
			try:
				a = list(map(lambda x: int(x, 16), a))
			except ValueError as a:
				raise RuntimeError("With -bin=LOAD,ADDR, LOAD and ADDR must be unprefixed hex numbers, but problem found: " + str(a))
			self.opt["bin"] = a[:2]
			a = set(a[1:])
			a.remove(self.opt["bin"][1])
			self.opt["bin"] += sorted(a)
		else:
			self.opt["bin"] = None
		if "toep" in self.opt:
			if "html" in self.opt or "info" in self.opt or "hex" in self.opt or "png" in self.opt:
				raise RuntimeError("Options -html, -info, -hex and -png are incompatible with -toep mode!")
			if not self.opt["out"]:
				raise RuntimeError("Specify an output file with -out in case of -toep mode!")
		if "png" in self.opt:
			if "html" in self.opt or "info" in self.opt or "hex" in self.opt:
				raise RuntimeError("Options -html, -info and -hex are incompatible with -png mode!")
			if not self.opt["out"]:
				raise RuntimeError("Specify an output file with -out in case of -png mode!")
		if self.opt["cset"] not in self.ascii:
			raise RuntimeError("Invalid EP-ASCII table \"" + self.opt["cset"] + "\", available ones: " + ", ".join(sorted(self.ascii)))
		self.ascii_tab = self.get_ascii_table(self.opt["cset"])
		# ascii_pc is a boolean map: if true, that char is the same as on PC, disassembler can encode it directly!
		self.ascii_pc = [a == self.get_unicode_pos(self.ascii_tab[a]) for a in range(len(self.ascii_tab))]
		self.ascii_pc[0xA3] = False # TODO FIXME! I have no idea why A3 is True, it shouldn't, this is a workaround!
		#print("ASCII_PC = " + " ".join([hex(a) + "[" + str(self.ascii_pc[a]) + "]" for a in range(len(self.ascii_pc)) if self.ascii_pc[a]]))
		self.forth_vlist = set([self.stringize(bytearray(a.encode("us-ascii")), conv_html = False) for a in self.forth_default_vlist])


	def list_options(self):

		return "Options:\n\n" + "\n".join([ "\t" + "-" + a + "\t" + self.OPTIONS[a] for a in sorted(self.OPTIONS)])

	def hexbyte(self, h):
		return hex(h).replace("x", "0")[-2:].upper()

	def hexdump(self, m):
		return " ".join([self.hexbyte(a) for a in m])

	def stringize(self, s, decor = None, padding = 0, conv_html = True):
		if isinstance(s, bytearray):
			padding -= len(s) # len must be got BEFORE conv (as after conv UTF8 more bytes for one EP byte)
			s = "".join([self.ascii_tab[a] for a in s])
		elif not isinstance(s, str):
			s = str(s)
			padding -= len(s)
		s = " " * padding + s
		if "html" in self.opt and conv_html:
			s = s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;") # replace(" ", "&nbsp;")
			if decor is not None:
				if decor[0] == "!":
					s = "<div class=\"" + decor[1:] + "\">" + s + "</div>"
				else:
					s = "<span class=\"" + decor + "\">" + s + "</span>"
		return s

	def __anchor(self, anchor):

		return "L" + str(self.exoshno) + "_" + str(anchor)

	def anchor(self, anchor):

		if "html" in self.opt and "nolinks" not in self.opt: return "<a name=\"" + self.__anchor(anchor) + "\"></a>"
		else: return ""

	def stringize_link(self, anchor, text, decor = None, padding = 0):

		text = self.stringize(text, decor = decor, padding = padding)
		if "html" in self.opt and "nolinks" not in self.opt: return "<a href=\"#" + self.__anchor(anchor) + "\">" + text + "</a>"
		else: return text

	def list_basic_line(self, line):
		#space = ("&nbsp;" if "html" in self.opt else " ")
		space = " "
		if "hex" in self.opt: o = self.stringize(self.hexdump(line), "hex") + "\n"
		else: o = ""
		p = line[0] | (line[1] << 8)
		o += self.anchor(p) + self.stringize(p, "linehead", padding = 5)
		o += space + space * line[2] * 2	# one space + possibly other spaces (*2) based on the nesting level
		p = 3
		while p < len(line):
			b = line[p]
			#print("Pars[" + hex(p) + "] = " + hex(b))
			if b < 0x20: # special sign
				o += self.stringize(self.signs[b], "sign")
				p += 1
			elif b < 0x60:
				b &= 31
				o += self.stringize(line[p + 1:p + 1 + b], "name")
				p += b + 1
			elif b == 0x80:
				b = line[p + 1]
				o += "\"" + self.stringize(line[p + 2:p + 2 + b], "string") + "\""
				p += b + 2
			elif b == 0x60:		# tokenized BASIC keyword
				b = line[p + 1]
				if b >= len(self.tokens):
					raise RuntimeError("Unknown token found: " + hex(b) + ", parsing state: " + o)
				b = self.tokens[b]
				o += self.stringize(b, "token")
				if b in self.untokenized:
					o += self.stringize(line[p + 2:], "data")
					break
				else:
					o += space # should I print a space after a basic token, always?
					p += 2
			elif b == 0xA2: # integer: BASIC line reference!
				b = line[p + 1] | (line[p + 2] << 8)
				o += self.stringize_link(b, " " + str(b), "number")
				p += 3
			elif b == 0xC2: # integer
				o += self.stringize(line[p + 1] | (line[p + 2] << 8), "number")
				p += 3
			elif b == 0xC6:	# float constant, or so ... EP seems to use a very interesting BCD-like float storing with some extra.
				#print("FLOAT")
				#print(hexdump(line[p + 1:p + 7]))
				s = (self.hexbyte(line[p + 5]) + self.hexbyte(line[p + 4]) + self.hexbyte(line[p + 3]) + self.hexbyte(line[p + 2]) + self.hexbyte(line[p + 1]))
				b = line[p + 6] - 0x3F
				if b > 0: s = s[0:b] + "." + s[b:]
				else: s = "." + "0" * -b + s
				o += self.stringize(s.rstrip("0").rstrip("."), "number")
				p += 7
			else:
				raise RuntimeError("Unknown byte found in the stream: " + hex(b) + ", parsing state: " + o)
		return o

	def show_info(self, show_exos_header = True):

		o = ""
		if not self.exoshno:
			o += "File name: " + self.file_name + " (" + str(self.file_size) + " bytes)\n"
			o += "Options: " + " ".join([a  + "=" + ("'" + str(self.opt[a]) + "'" if isinstance(self.opt[a], str) else str(self.opt[a])) for a in sorted(self.opt)]) + "\n"
			o += "EP-ASCII: " + self.opt["cset"] + "\n"
		if show_exos_header:
			o += "EXOS header pos/serial: #" + str(self.exoshno) + " at " + hex(self.pos) + "\n"
			o += "EXOS type/header: " + hex(self.exostype) + " [" + self.hexdump(self.exosheader) + "]\n"
			o += "EXOS type determined: " + self.known_exostypes[self.exostype][0] + "\n"
		return o

	def list_basic_multiple_header(self, listparam):

		if self.list_basic_program(None) > 0:
			raise RuntimeError("Multiple IS-BASIC module should be zero length!")

	def list_basic_program(self, listparam):

		ml = self.exosheader[2] | (self.exosheader[3] << 8)
		if ml == 0:
			if "info" in self.opt:
				self.output += self.stringize("Program: " + hex(self.exosheader[4]) + ", EMPTY MODULE.", "!stat")
			return 0
		nextpos = self.pos + ml
		if "info" in self.opt:
			self.output += self.stringize("Program: " + hex(self.exosheader[4]) + ", size: " + hex(ml) + " bytes.", "!stat")
		while True:
			ll = self.prg[self.pos]
			#print("linelength=" + str(ll))
			if ll == 0: break
			if self.prg[self.pos + ll - 1] != 0: raise RuntimeError("Invalid BASIC line, not terminated with zero")
			try:
				self.output += self.list_basic_line(self.prg[self.pos + 1:self.pos + ll - 1]) + "\n"
			except IndexError:
				raise RuntimeError("Unexpected end of BASIC stream or damaged?")
			self.pos += ll
		self.pos += 1
		if nextpos != self.pos:
			raise RuntimeError("Size inconsistency: header specified IS-BASIC position does not match with the value after parsing! Damaged file?")
		return ml

	def show_forth_word(self, tok, ltok):

		if self.forth_string:
			if tok[-1] == '"':
				self.forth_string = 0
				return self.stringize(tok[:-1], "string") + '"'
			self.forth_string += 1
			return self.stringize(tok, "string")
		else:
			if tok[-1] == '"': self.forth_string = 1
		if tok == "(":
			self.forth_comment = True
		if self.forth_comment:
			if tok == ")": self.forth_comment = False
			return self.stringize(tok, "comment")
		if ltok == ":":
			self.vlist.add(tok)
			return self.anchor(tok) + self.stringize(tok, "name")
		if tok in self.vlist:
			return self.stringize_link(tok, tok, "name")
		if tok in self.forth_vlist:
			if tok == "DECIMAL": self.forth_nummode = "0123456789"
			elif tok == "BINARY": self.forth_nummode = "01"
			elif tok == "HEX": self.forth_nummode = "0123456789ABCDEF"
			elif tok == "OCTAL": self.forth_nummode = "01234567"
			return self.stringize(tok, "token")
		if all([a in self.forth_nummode for a in tok]): # it seems forth allows redefine even numbers, so it's the last 'if'
			return self.stringize(tok, "hex")
		return tok

	def list_forth_line(self, line):

		tok, ltok, o = bytearray(), None, ""
		self.forth_comment, self.forth_string = False, 0
		self.forth_nummode = "0123456789"
		for c in line:
			if c in (32, 9):
				if tok:
					tok = self.stringize(tok, conv_html = False)
					o += self.show_forth_word(tok, ltok)
					ltok, tok = tok, bytearray()
				o += self.stringize(bytearray([c]), decor = ("string" if self.forth_string > 1 else None))
			else:
				tok += bytearray([c])
		if tok:
			o += self.show_forth_word(self.stringize(tok, conv_html = False), ltok)
		return o

	def list_forth_program(self, listparam):

		nbufs, self.vlist, cbuf = self.exosheader[2], set(), 0
		space_byte, crlf_bytes = bytearray([0x20]), bytearray([13, 10])
		while cbuf < nbufs:
			buf, self.pos, cbuf = self.prg[self.pos : self.pos + 1026], self.pos + 1026, cbuf + 1
			if len(buf) != 1026:
				raise RuntimeError("Cannot extract the next buffer (1024+2 bytes) from the stream, got only " + str(len(buf)) + " byte(s).")
			bufid = buf[0] | (buf[1] << 8)
			if "hex" in self.opt:
				self.output += self.stringize(self.hexdump(buf[:2]), "hex") + "\n"
			buf = buf[2:].rstrip(space_byte)
			lines = buf.rstrip(crlf_bytes).split(crlf_bytes)
			if "info" in self.opt:
				self.output += self.stringize("New forth buffer #"  + str(bufid) + " (" + str(cbuf) + " of " + str(nbufs) + ")\n" +
				"Used bytes: " + str(len(buf)) + ", (usefull) lines: " + str(len(lines))
				, "!stat") + "\n"
			for a in lines:
				if "hex" in self.opt:
					self.output += self.stringize(self.hexdump(a + crlf_bytes), "hex") + "\n"
				self.output += self.list_forth_line(a) + "\n"
		if "info" in self.opt:
			self.output += self.stringize("(Re-)defined forth words: " + str(len(self.vlist)), "!stat")
		if self.pos == len(self.prg):
			self.stop = "No other EXOS header after IS-FORTH program"
		elif self.pos > len(self.prg):
			raise RuntimeError("Truncated IS-FORTH/EXOS stream?!")

	def list_editor_file(self, listparam):

		try:
			while True:
				p2 = self.pos + 3
				while self.prg[p2] >= 32: p2 += 1
				if "hex" in self.opt: self.output += self.stringize(self.hexdump(self.prg[self.pos : p2 + 1]), "hex") + "\n"
				self.output += self.stringize(self.prg[self.pos + 3 : p2], "string") + "\n"
				#print("[" + self.stringize(self.prg[self.pos + 3 : p2], "string") + "]")
				self.pos = p2 + 1
				if self.prg[p2] == 2: break
		except IndexError:
			raise RuntimeError("Unexpected end of EDITOR BUFFER stream or damaged?")

	def disasm_program(self, listparam):

		size_exos = self.exosheader[2] | (self.exosheader[3] << 8)
		size_file = len(self.prg) - self.pos
		if size_exos != size_file:
			raise RuntimeError("[EXOS header][type=" + str(self.exostype) + "] File size information on EXOS header (" + str(size_exos) + ") mismatch with amount of read bytes (" + str(size_file) + ")")
		o = self.analyze_ml_prg(
			prg = self.prg[self.pos:],
			size = size_file,
			load = listparam["start"],
			start = [listparam["start"]],
			html = "html" in self.opt,
			debug = "hex" in self.opt,
			dasm = "dasm" in self.opt,
			preprint = ("\tORG " + hex(listparam["start"] - 16) + "\n\tDB 0x" + self.hexdump(self.exosheader).replace(" ", ",0x") + " ; EXOS header" if "dasm" in self.opt else None)
		)
		if "info" in self.opt:
			self.output += self.stringize("\n".join(self.dasminfo), "!stat") + "\n"
		self.output += o
		self.stop = "After type-" + str(self.exostype) + " header"

	def end_of_module(self, listparam):

		self.stop = "End-of-module header got (type=10)"

	def not_supported(self, listparam = None):

		raise RuntimeError("Sorry, but this EXOS type is not supported")

	def list_heass_file(self, listparam):

		self.stop = "After HEASS file"
		first = "Filename: "
		while self.pos < len(self.prg) and self.prg[self.pos] != 0xFF:
			l = self.prg[self.pos]
			if "hex" in self.opt:
				self.output += self.stringize(self.hexdump(self.prg[self.pos : self.pos + l + 1]), "hex") + "\n"
			self.pos += 1
			while l:
				c = self.prg[self.pos]
				self.pos += 1
				if c > 0x9F:
					c -= 0x9F
					if first:
						first += " " * c
					else:
						self.output += " " * c
				else:
					if first:
						first += self.stringize(bytearray([c]))
					else:
						self.output += self.stringize(bytearray([c]))
				l -= 1
			if first:
				if "info" in self.opt:
					self.output += self.stringize(first, "!stat")
				first = False
				self.pos += 10
			self.output += '\n'

	def image_init(self, xsize, ysize):

		# Import needed modules
		global StringIO, b64encode, Image
		try:
			from StringIO import StringIO
		except Exception as a:
			raise RuntimeError("Cannot import StringIO python module: " + str(a))
		try:
			from base64 import b64encode
		except Exception as a:
			raise RuntimeError("Cannot import base64 python module: " + str(a))
		try:
			from PIL import Image
		except Exception as a:
			raise RuntimeError("Cannot import Image module from PIL (Python Imaging Library). You must install it on your computer! Msg: " + str(a))
		# Create image
		img = Image.new("P", (xsize, ysize))
		# Palette creation (EP/Nick palette in RGB colour space)
		pal = []
		for a in range(256):
			pal.append(int((((a &  1) << 10) | ((a &  8) << 6) | ((a &  64) << 2)) / (1792 / 255.0)))
			pal.append(int((((a &  2) <<  9) | ((a & 16) << 5) | ((a & 128) << 1)) / (1792 / 255.0)))
			pal.append(int((                   ((a &  4) << 7) | ((a &  32) << 3)) / ( 768 / 255.0)))
		img.putpalette(pal)
		return img


	def image_output(self, img, resize = None):

		if resize is not None:
			img = img.resize(resize)
		imd = StringIO()
		img.save(imd, "GIF")
		if "savepic" in self.opt:
			f = self.opt["savepic"]
			if f == "":
				img.show()
			else:
				try:
					img.save(f, "GIF")
				except IOError as a:
					raise RuntimeError("Cannot save image (-savepic was specified) because of: " + str(a))
				self.output += self.stringize(
					"Image has been saved as filename: " + f,
					"!stat"
				)
		if "html" not in self.opt:
			self.output += "\n\nIMAGE can be shown only in html (-html) mode!\nYou can use -savepic=filename to save the image as well\n"
		else:
			self.output += '<img alt="image data" src="data:image/GIF;base64,' + b64encode(imd.getvalue()) + '">'


	def show_iview_image(self, listparam):

		####raise RuntimeError("IVIEW image is not supported yet :-(")
		#if self.exosheader[0xA] != 0:
		#	raise RuntimeError("Currently only uncompressed IVIEW files are supported and this is a compressed one.")
		self.stop = "After IVIEW image"
		if "info" in self.opt:
			self.output += self.stringize(
				"Change info per lines (0=always same): " +
				"vidmod=" + str(self.exosheader[2]) +
				", fixbias=" + str(self.exosheader[3]) +
				", palette=" + str(self.exosheader[4]) + "\n" +
				"Interlace: 2xvres=" + str(bool(self.exosheader[5] & 128)) +
				", viddat=" + str(bool(self.exosheader[5] & 16)) +
				", attrib=" + str(bool(self.exosheader[5] & 8)) +
				", palette=" + str(bool(self.exosheader[5] & 4)) +
				", fixbias=" + str(bool(self.exosheader[5] & 2)) +
				", vidmode=" + str(bool(self.exosheader[5] & 1)) + "\n" +
				"Height: " + str(self.exosheader[6] | (self.exosheader[7] << 8))
				,"!stat"
			)


	def show_vload_image(self, listparam):

		# Some defaults
		gmodes = {0: "hwtext", 1: "pixel", 2: "swtext", 5: "lpixel", 15: "attribute"}
		palette_indexed = [0x00, 0x24, 0x49, 0xDB, 0x92, 0x2D, 0x36, 0xFF, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07]
		# Extract / form important information
		gmod, cmod = self.exosheader[2], self.exosheader[3] & 3
		lx, ly = self.exosheader[4], self.exosheader[5]
		lxp, lyp = lx << (3 - cmod), ly * 9
		bs, cols = lx * lyp, 1 << (1 << cmod)
		lxo, lyo = lxp << cmod, lyp << 1
		# Check supported types / size
		if gmod not in gmodes:
			raise RuntimeError("[EXOS type-0Bh] Unknown GFX mode descriptor byte: " + hex(gmod))
		if gmod != 1:
			raise RuntimeError("[EXOS type-0Bh] Currently, only PIXEL mode is supported, and this is " + gmodes[gmod] + " :-(")
		if self.pos + bs > self.file_size:
			raise RuntimeError("[EXOS type-0Bh] Not enough bytes for the image. Bad/damaged file?")
		# Create image
		img = self.image_init(lxp, lyp)
		pix = img.load()
		# Fill image data
		for y in range(lyp):
			for x in range(lx):
				b, self.pos = self.prg[self.pos], self.pos + 1
				if cmod == 3:
					# 256 colour mode (not palette indexed mode!)
					pix[x, y] = b
				elif cmod == 2:
					# 16 colour mode
					pix[ x << 1,      y] = palette_indexed[((b & 2) << 2) | ((b & 32) >> 3) | ((b & 8) >> 2) | ((b & 128) >> 7)]
					pix[(x << 1) + 1, y] = palette_indexed[((b & 1) << 3) | ((b & 16) >> 2) | ((b & 4) >> 1) | ((b &  64) >> 6)]
				elif cmod == 1:
					# 4 colour mode
					pix[ x << 2,      y] = palette_indexed[((b & 8) >> 2) | ((b & 128) >> 7)]
					pix[(x << 2) + 1, y] = palette_indexed[((b & 4) >> 1) | ((b &  64) >> 6)]
					pix[(x << 2) + 2, y] = palette_indexed[ (b & 2)       | ((b &  32) >> 5)]
					pix[(x << 2) + 3, y] = palette_indexed[((b & 1) << 1) | ((b &  16) >> 4)]
				else:
					# 2 colour mode
					for z in range(8):
						pix[(x << 3) + z, y] = palette_indexed[(b >> (7 - z)) & 1]
		# Colour usage stat
		z = set()
		for y in range(lyp):
			for x in range(lxp):
				z.add(pix[x, y])
		# Information
		if "info" in self.opt:
			self.output += self.stringize(
				"Image (char-units): " + str(lx) + " x " + str(ly) + "\n" +
				"Image (bytes): " + str(bs) + " (" + str(lx) + " x " + str(lyp) + ")\n" +
				"Image (pixels): " + str(lxp) + " x " + str(lyp) + "\n" +
				"Image gfx mode: " + gmodes[gmod] + " [id=" + hex(gmod) + "]\n" +
				"Image colour mode: " + str(cols) + "col [id=" + hex(cmod) + "]\n" +
				"Image scaled to: " + str(lxo) + " x " + str(lyo) + "\n" +
				"Default indexed palette: " + ("NON-INDEXED MODE" if cmod == 3 else self.hexdump(palette_indexed[:cols])) + "\n" +
				"Really used number of colours: " + str(len(z))
				,
				"!stat"
			) + "\n"
		# Finally, create the output!
		self.image_output(img, (lxo, lyo))


	def list_exos_rom(self):

		if "info" in self.opt:
			self.output += self.stringize(self.show_info(show_exos_header = False) + "\nSpecial mode: EXOS_ROM listing", "!stat")
		o = self.analyze_ml_prg(
			prg = self.prg,
			size = self.file_size,
			load = 0xC000,
			start = [0xC00A],
			html = "html" in self.opt,
			debug = "hex" in self.opt,
			dasm = "dasm" in self.opt,
			preprint = ("\tORG 0xC000" if "dasm" in self.opt else "")
		)
		if "info" in self.opt:
			self.output += self.stringize("\n".join(self.dasminfo), "!stat") + "\n"
		self.output += o

	def parse_file(self, name):

		self.file_name = name
		with open(name, "rb") as f: self.prg = bytearray(f.read(1048577))
		if len(self.prg) > 1048576: raise RuntimeError("Too long file (> 1M).")
		if len(self.prg) < 16 and self.opt["bin"] is None: raise RuntimeError("Too short file (< 16 bytes).")
		self.file_size = len(self.prg)
		self.output = ""
		self.exostype = 0
		self.exoshno = 0
		if "html" in self.opt: self.output += self.html_head + self.html_listing_head + "<pre>"
		if self.prg[:8] == bytearray("EXOS_ROM".encode("us-ascii")) and not self.opt["bin"]:
			if self.file_size > 0x4000:
				raise RuntimeError("Input seems to be an EXOS ROM, but currently only <=16K sized images are supported.")
			self.list_exos_rom()
			self.stop = "End of EXOS_ROM image"
			self.parse_file_add_tail()
			return
		#self.exoshno = 0
		self.pos = 0
		#if "html" in self.opt: self.output += self.html_head + self.html_listing_head + "<pre>"
		if self.opt["bin"]:
			if "info" in self.opt:
				self.output += self.stringize(self.show_info(show_exos_header = False) + "\nSpecial mode: BINARY image listing", "!stat")
			o = self.analyze_ml_prg(
				prg = self.prg,
				size = self.file_size,
				load = self.opt["bin"][0],
				start = self.opt["bin"][1:],
				html = "html" in self.opt,
				debug = "hex" in self.opt,
				dasm = "dasm" in self.opt,
				preprint = (("\tORG " + hex(self.opt["bin"][1])) if "dasm" in self.opt else "")
			)
			if "info" in self.opt:
				self.output += self.stringize("\n".join(self.dasminfo), "!stat") + "\n"
			self.output += o
			self.stop = "Binary image mode"  # cause to stop parsing "EXOS header" before it can begin it :)
		else:
			self.stop = False
		while not self.stop:
			if self.exoshno:
				self.output += "\n"
				if self.pos == len(self.prg):
					self.stop = "No new EXOS header found at the current position (end of file)"
					break
			if self.file_size - self.pos < 16: raise RuntimeError("[EXOS header] Truncated header, EOF; start of header at " + hex(self.pos))
			if self.prg[self.pos]: raise RuntimeError("[EXOS header] First byte of header is non-zero at " + hex(self.pos) + ". Not an EXOS file?")
			if self.prg[self.pos + 15]: raise RuntimeError("[EXOS header] Last byte of header is non-zero at " + hex(self.pos + 15) + ". Not an EXOS file?")
			self.exostype = int(self.prg[self.pos + 1])
			if self.exostype == 0: raise RuntimeError("[EXOS header] Second byte is zero at " + hex(self.pos + 1) + ". Not an EXOS file?")
			self.exosheader = self.prg[self.pos : self.pos + 16]
			#if self.exostype not in self.known_exostypes or self.known_exostypes[self.exostype][2] == False:
			if self.exostype not in self.known_exostypes:
				#self.not_supported()
				raise RuntimeError("[EXOS header] Unsupported type " + str(self.exostype) + " at " + hex(self.pos + 1))
			if "info" in self.opt:
				self.output += self.stringize(self.show_info(), "!stat") + "\n"
			self.pos += 16 # ok, now after the header ...
			self.known_exostypes[self.exostype][1](self, self.known_exostypes[self.exostype][2])
			self.exoshno += 1 # next header
		self.parse_file_add_tail()

	def parse_file_add_tail(self):

		if "info" in self.opt: self.output += "\n" + self.stringize("End of FILE parsing, because: " + self.stop, "!stat") + "\n"
		if "html" in self.opt: self.output += "</pre>" + self.html_tail
		if "png" in self.opt: self.output = self.create_png(self.output)
		if self.opt["out"]:
			with open(self.opt["out"], "w") as f:
				f.write(self.output)
		else:
			print(str(self.output))
		if self.exostype == 10 and self.pos != self.file_size:
			print("WARNING: extra garbage at the end of file after header-10?!")


	def create_png(self, text):

		try:
			import PIL
		except ImportError as e:
			raise RuntimeError("For -png support you must have \"PIL\" module for Python installed\n" + str(e))
		raise RuntimeError("Sorry, PNG support is not ready yet :-(")


	known_exostypes = {
		0x01: ("IS-FORTH program", list_forth_program, None),
		0x03: ("IS-BASIC multiple header", list_basic_multiple_header, None),
		0x04: ("IS-BASIC program (single)", list_basic_program, None),
		0x05: ("User program (ML)", disasm_program, {"start": 0x0100}),
		0x06: ("Absolute system extension", disasm_program, {"start": 0xC00A}),
		0x08: ("Editor document file (eg: WP)", list_editor_file, None),
		0x0A: ("End of module", end_of_module, None),
		0x0B: ("VLOAD image", show_vload_image, None),
		0x49: ("IVIEW image", show_iview_image, None),
		0x80: ("HEASS source", list_heass_file, None),
	}



if __name__ == "__main__":

	ep = EP128FileLister()
	try:
		if len(sys.argv) < 2:
			die("Usage: " + sys.argv[0] + " [options] filename\n\n" + ep.list_options() + "\n\n-db or -chset without any other opts/name: dumps internal DBs or\n\tcharacter sets as HTML")
		elif len(sys.argv) == 2 and sys.argv[1] == "-db":
			print(ep.dump_db())
		elif len(sys.argv) == 2 and sys.argv[1] == "-chset":
			print(ep.dump_chset())
		elif len(sys.argv) == 2 and sys.argv[1] == "-version":
			print(ep.VERSION)
		elif len(sys.argv) == 2 and sys.argv[1] == "-z80ops":
			print(ep.dump_z80opcodes())
		else:
			ep.set_options(sys.argv[1:-1])
			ep.parse_file(sys.argv[-1])
	except KeyboardInterrupt:
		die("Aborting (CTRL-C was pressed?)")
	except IOError as e:
		die("File I/O error " + str(e))
	except RuntimeError as e:
		#raise
		die(str(e))
	sys.exit(0)
